/**
 * Copyright © 2010-2011 Olivier Cailloux
 *
 *     This file is part of JLP.
 *
 *     JLP is free software: you can redistribute it and/or modify it under the
 *     terms of the GNU Lesser General Public License version 3 as published by
 *     the Free Software Foundation.
 *
 *     JLP is distributed in the hope that it will be useful, but WITHOUT ANY
 * 	WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * 	FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for
 * 	more details.
 *
 *     You should have received a copy of the GNU Lesser General Public License
 *     along with JLP. If not, see <http://www.gnu.org/licenses/>.
 */
package org.decisiondeck.jlp.problem;

import java.util.Set;

import org.decisiondeck.jlp.LpConstraint;


import com.google.common.base.Preconditions;
import com.google.common.collect.Sets;

/**
 * Utilities methods (e.g. views, copies) related to a {@link LpProblem}.
 * 
 * @author Olivier Cailloux
 * 
 */
public class LpProblemUtils {
    static public <T> LpProblem<T> getImmutableCopy(LpProblem<T> source) {
	return new LpProblemImmutable<T>(source);
    }

    static public <T> LpProblem<T> getReadView(LpProblem<T> delegate) {
	return new LpProblemReadView<T>(delegate);
    }

    /**
     * Retrieves a long description, with line breaks, of the given problem.
     * 
     * @param <T>
     *            the type of the variables in the problem.
     * @param problem
     *            not <code>null</code>.
     * @return not <code>null</code>, not empty.
     */
    static public <T> String getAsLongString(LpProblem<T> problem) {
	Preconditions.checkNotNull(problem);
	String N = System.getProperty("line.separator");
	final String name = problem.getName().equals("") ? "" : " " + problem.getName();
	String s = "Problem" + name + N;

	if (!problem.getObjective().isEmpty()) {
	    s += problem.getObjective().getDirection() + N;
	    s += " " + problem.getObjective().getFunction() + N;
	} else {
	    s += "Find one solution" + N;
	}
	s += "Subject To" + N;
	for (LpConstraint<T> constraint : problem.getConstraints()) {
	    s += "\t" + constraint + N;
	}
	s += "Bounds" + N;
	for (T variable : problem.getVariables()) {
	    Number lb = problem.getVarLowerBound(variable);
	    Number ub = problem.getVarUpperBound(variable);

	    if (lb != null || ub != null) {
		s += "\t";
		if (lb != null) {
		    s += lb + " <= ";
		}
		s += variable;
		if (ub != null) {
		    s += " <= " + ub;
		}
		s += N;
	    }
	}

	s += "Variables" + N;
	for (T variable : problem.getVariables()) {
	    s += "\t" + variable + " " + problem.getVarType(variable) + N;
	}

	return s;

    }

    /**
     * Retrieves all the variables in the given problem that have a given type.
     * 
     * @param <T>
     *            the class of the variables.
     * @param problem
     *            not <code>null</code>.
     * @param type
     *            not <code>null</code>.
     * @return not <code>null</code>.
     */
    static public <T> Set<T> getVariables(LpProblem<T> problem, LpVariableType type) {
	Preconditions.checkNotNull(problem);
	Preconditions.checkNotNull(type);
	final Set<T> searched = Sets.newHashSet();
	final Set<T> variables = problem.getVariables();
	for (T variable : variables) {
	    final LpVariableType varType = problem.getVarType(variable);
	    if (varType == type) {
		searched.add(variable);
	    }
	}
	return searched;
    }

    /**
     * <p>
     * Retrieves a view of the given problem that has its own name. The name of the underlying problem is ignored even
     * if the provided new name is empty.
     * </p>
     * <p>
     * This method is provided for efficiency reasons, as copying the problem to a new problem in order to change its
     * name may be long for big problems. The provided view is writable, writing to it delegates to the underlying
     * problem except that changing its name change only its internal name, not the name of the delegate problem.
     * </p>
     * 
     * @param <T>
     *            the type of the variables in the problem.
     * @param problem
     *            the underlying problem.
     * @param problemName
     *            the new problem name, <code>null</code> or empty for no name.
     * @return not <code>null</code>.
     */
    static public <T> LpProblem<T> getViewWithName(LpProblem<T> problem, String problemName) {
	final LpProblemOwnName<T> view = new LpProblemOwnName<T>(problem);
	view.setName(problemName);
	return view;
    }
}
